@@ -17,7 +17,7 @@ sub connect
   }
 
 sub blob {
-    return "VARCHAR2(4000)";
+    return "CLOB";
 }
 
 sub date {
@@ -28,6 +28,23 @@ sub bool {
     return "INT(1)";
 }
 
+# Oracle--
+sub from_date {
+    $_[1];
+    #print STDERR "Converting FROM $_[1]\n";
+    #(my $date = $_[1]) =~ s{ }{T};
+    #$date;
+ }
+sub to_date {
+    $_[1];
+    #print STDERR "Converting TO $_[1]\n";
+    #(my $date = $_[1]) =~ s{T}{ };
+    #$date;
+}
+
+sub from_blob { $_[1] }
+sub to_blob { $_[1] }
+
 package Tangram::Oracle::Storage;
 
 use Tangram::Storage;
@@ -43,7 +60,7 @@ sub connect
     # Oracle doesn't really have a default date format (locale
     # dependant), so adjust it to use ISO-8601.
     $self->{db}->do
-	("ALTER SESSION SET NLS_DATE_FORMAT='YYYYMMDDHH24:MI:SS'");
+	("ALTER SESSION SET NLS_DATE_FORMAT='YYYY-MM-DD\"T\"HH24:MI:SS'");
     $self->{db}->do
 	("ALTER SESSION SET CONSTRAINTS = DEFERRED");
     $self->{db}->{RaiseError} = 1;
