@@ -1,24 +1,40 @@
-package Tangram::Storage;
+# (c) Sound Object Logic 2000-2001
 
 use strict;
 
-use Tangram::Storage::Statement;
-
+package Tangram::Storage;
 use DBI;
 use Carp;
-
-use Tangram::Util qw(pretty);
-use Scalar::Util qw(weaken refaddr);
+use Tangram::Core;
 
 use vars qw( %storage_class );
 
+BEGIN {
+    eval 'use Scalar::Util qw()';
+    if ($@) {
+	*Tangram::refaddr = sub { (shift) + 0 };
+	eval 'use WeakRef';
+	if ($@) {
+	    *Tangram::weaken = sub { };
+	    $Tangram::no_weakrefs = 1;
+	} else {
+	    *Tangram::weaken = \&WeakRef::weaken;
+	    $Tangram::no_weakrefs = 0;
+	}
+    } else {
+	*Tangram::weaken = \&Scalar::Util::weaken;
+	*Tangram::refaddr = \&Scalar::Util::refaddr;
+	$Tangram::no_weakrefs = 0;
+    }
+    *pretty = *Tangram::Core::pretty;
+}
+
 sub new
 {
     my $pkg = shift;
     return bless { @_ }, $pkg;
 }
 
-# XXX - not tested by test suite
 sub schema
 {
     shift->{schema}
@@ -41,8 +57,6 @@ sub split_id
 
 use Scalar::Util qw(looks_like_number);
 
-# Given a row's ID and a class's ID
-# Computes its OID and returns it
 sub combine_ids
   {
 	my $self = shift;
@@ -68,7 +82,6 @@ sub to_dbms
 	return $self->{driver}->to_dbms(@_);
     }
 
-# XXX - not tested by test suite
 sub get_sequence {
     my $self = shift;
     my $sequence_name = shift;
@@ -77,14 +90,11 @@ sub get_sequence {
     # the only database that has a non-trivial sequence sql fragment
     # also doesn't use " FROM DUAL"
     my $query = $self->sequence_sql($sequence_name).$self->from_dual;
-    my ($id) = (map { @{$_} }
-		map { @{$_} }
-		$self->{db}->selectall_arrayref($query));
+    my ($id) = map { @$_ } $self->{db}->selectall_arrayref($query);
 
     return $id;
 }
 
-# XXX - not tested by test suite
 sub sequence_sql
     {
 	my $self = shift;
@@ -92,14 +102,6 @@ sub sequence_sql
 	return $self->{driver}->sequence_sql(shift);
     }
 
-# XXX - not tested by test suite
-sub limit_sql {
-    my $self = shift;
-
-    my $driver = $self->{driver} or confess "no driver";
-    return $self->{driver}->limit_sql(@_);
-}
-
 sub _open
   {
     my ($self, $schema) = @_;
@@ -117,7 +119,6 @@ sub _open
 	  local $dbh->{PrintError} = 0;
 	  my $control;
 	  if ( $schema->{sql}{oid_sequence} ) {
-	      # XXX - not tested by test suite
 	      $control = "dummy";
 	  } else {
 	      $control = $dbh->selectall_arrayref
@@ -132,18 +133,17 @@ sub _open
 		$self->{import_id} = sub { shift() . sprintf("%0$self->{cid_size}d", shift()) };
 		$self->{export_id} = sub { substr shift(), 0, -$self->{cid_size} };
 	  } else {
-	      # XXX - layout1
-	      $self->{class_col} = 'classId';
-	      $self->{layout1} = 1;
-	      $self->{import_id} = sub { shift() };
-	      $self->{export_id} = sub { shift() };
+		$self->{class_col} = 'classId';
+		$self->{layout1} = 1;
+		$self->{import_id} = sub { shift() };
+		$self->{export_id} = sub { shift() };
 	  }
 	}
 
 	my %id2class;
 
 	if ($self->{layout1}) {
-	    # XXX - layout1
+	  # compatibility with version 1.x
 	  %id2class = map { @$_ } @{ $self->{db}->selectall_arrayref("SELECT classId, className FROM $schema->{class_table}") };
 	} else {
 	  my $classes = $schema->{classes};
@@ -166,32 +166,25 @@ sub _open
 		}
 	  }
 	  if ($id) {
-	    $self->{ids}{refaddr($obj)} = $id;
+	    $self->{ids}{Tangram::refaddr($obj)} = $id;
 	  } else {
-	    delete $self->{ids}{refaddr($obj)};
+	    delete $self->{ids}{Tangram::refaddr($obj)};
 	  }
 	};
 
     $self->{get_id} = $schema->{get_id} || sub {
 	  my $obj = shift or warn "no object passed to get_id";
-	  ref $obj or return undef;
-	  my $address = refaddr($obj)
-	      or do { warn "Object $obj has no refaddr(?)";
-		      return undef };
+	  my $address = Tangram::refaddr($obj);
 	  my $id = $self->{ids}{$address};
-	  # refaddr's can be re-used, but weakrefs are magic :-)
-	  if ( $id and !defined $self->{objects}{$id} ) {
-	      delete $self->{ids}{$address};
-	      delete $self->{objects}{$id};
-	      $id = undef;
-	  } elsif ( $id and refaddr($self->{objects}{$id}) != $address ) {
-	      delete $self->{ids}{$address};
-	      $id = undef;
-	  }
 	  if ($Tangram::TRACE && ($Tangram::DEBUG_LEVEL > 2)) {
 		print $Tangram::TRACE "Tangram: $obj is ".($id?"oid $id" : "not in storage")."\n";
 	  }
-	  return $id;
+	  return undef unless $id;
+	  # refaddr's can be re-used, but weakrefs are magic :-)
+	  return $id if defined($self->{objects}{$id});
+	  delete $self->{ids}{$address};
+	  delete $self->{objects}{$id};
+	  return undef;
 	};
 
     return $self;
@@ -206,7 +199,6 @@ sub alloc_table
 	    : ++$self->{table_top};
 }
 
-# XXX - not reached by test suite
 sub free_table
 {
     my $self = shift;
@@ -231,7 +223,6 @@ sub open_connection
     return $db;
 }
 
-# XXX - not reached by test suite
 sub close_connection
   {
     # private - close read connection to DB unless it's the default one
@@ -269,14 +260,12 @@ sub my_cursor
     return $cursor;
 }
 
-# XXX - not reached by test suite
 sub select_data
 {
     my $self = shift;
-    Tangram::Expr::Select->new(@_)->execute($self, $self->open_connection());
+    Tangram::Select->new(@_)->execute($self, $self->open_connection());
 }
 
-# XXX - not reached by test suite
 sub selectall_arrayref
 {
     shift->select_data(@_)->fetchall_arrayref();
@@ -285,7 +274,7 @@ sub selectall_arrayref
 sub my_select_data
 {
     my $self = shift;
-    Tangram::Expr::Select->new(@_)->execute($self, $self->{db});
+    Tangram::Select->new(@_)->execute($self, $self->{db});
 }
 
 my $psi = 1;
@@ -304,10 +293,9 @@ sub prepare
 *prepare_update = \&prepare;
 *prepare_select = \&prepare;
 
-# XXX - lots of options here not tested by test suite
 sub make_id
   {
-    my ($self, $class_id, $o) = @_;
+    my ($self, $class_id) = @_;
 
     # see if the class has its own ID generator
     my $cname = $self->{id2class}{$class_id};
@@ -315,23 +303,19 @@ sub make_id
 
     my $id;
     if ( $classdef->{make_id} ) {
-	$id = $classdef->{make_id}->($class_id, $self, $o);
+	$id = $classdef->{make_id}->($class_id, $self);
 	print $Tangram::TRACE "Tangram: custom per-class ($cname) make ID function returned ".(pretty($id))."\n" if $Tangram::TRACE;
     } elsif ( $classdef->{oid_sequence} ) {
-	eval { $id = $self->get_sequence($classdef->{oid_sequence}) };
-	die "Failed to get sequence for Class `$cname'; $@" if $@;
+	$id = $self->get_sequence($classdef->{oid_sequence});
     }
 
     # maybe the entire schema has its own ID generator
     if ( !defined($id) and $self->{schema}{sql}{make_id} ) {
-	$id = $self->{schema}{sql}{make_id}->($class_id, $self, $o);
+	$id = $self->{schema}{sql}{make_id}->($class_id, $self);
 	print $Tangram::TRACE "Tangram: custom schema make ID function returned "
 	    .(pretty($id))."\n" if $Tangram::TRACE;
-    } elsif ( !defined($id) &&
-	      (my $seq = $self->{schema}{sql}{oid_sequence}) ) {
-	eval { $id = $self->get_sequence($seq) };
-	die "Failed to get sequence for Class `$cname' via fallback $seq; $@"
-	    if $@;
+    } elsif ( my $seq = $self->{schema}{sql}{oid_sequence} ) {
+	$id = $self->get_sequence($seq);
     }
     if (defined($id)) {
 	return $self->combine_ids($id, $class_id);
@@ -351,8 +335,6 @@ sub make_id
 	  return sprintf "%d%0$self->{cid_size}d", $id, $class_id;
 	}
 
-    # XXX - layout1
-
 	# ------------------------------
 	# compatibility with version 1.x
 
@@ -422,7 +404,6 @@ sub unknown_classid
 
 {
     no strict 'refs';
-# Given a class name ('Foo::Bar'), returns its Class ID.
 sub class_id
 {
     my $self = shift;
@@ -514,7 +495,6 @@ sub tx_commit
 	  delete $self->{alloc_id};
 	}
 	
-    # XXX - layout1
 	# compatibility with version 1.x
 	# ------------------------------
     
@@ -651,7 +631,7 @@ sub _insert
 
     my $class = $self->{schema}->classdef($class_name);
 
-    my $id = $self->make_id($classId, $obj);
+    my $id = $self->make_id($classId);
 
     $self->welcome($obj, $id);
     $self->tx_on_rollback( sub { $self->goodbye($obj, $id) } );
@@ -685,8 +665,7 @@ sub _insert
 		my @sql = $engine->get_insert_statements($class);
 		printf $Tangram::TRACE ">-\n%s\n".(@{$fields[$i]}?"-- with:\n    /* (%s) */\n":"%s")."...\n",
 		$sql[$i],
-		join(', ', map { defined($_)?$dbh->quote($_):"NULL" }
-		     @state[ @{ $fields[$i] } ] )
+		join(', ', map { $_ || 'NULL' } @state[ @{ $fields[$i] } ] )
 	  }
 
 	  my $sth = $sths->[$i];
@@ -764,8 +743,7 @@ sub _update
 		my @sql = $engine->get_update_statements($class);
 		printf $Tangram::TRACE ">-\n%s\n-- with\n    /* (%s) */\n...\n",
 		$sql[$i],
-		join(', ', map { defined($_)?$dbh->quote($_):"NULL" }
-		     @state[ @{ $fields[$i] } ] )
+		join(', ', map { $_ || 'NULL' } @state[ @{ $fields[$i] } ] )
 	  }
 
 	  my $sth = $sths->[$i];
@@ -777,7 +755,6 @@ sub _update
 #############################################################################
 # save
 
-# XXX - not documented / tested
 sub save
   {
     my $self = shift;
@@ -873,28 +850,6 @@ sub defer
     push @{$self->{defered}}, $action;
 }
 
-# Given a class' name and a row's ID (or more than one,)
-# computes the OIDs and returns them.
-# XXX - not tested by test suite
-sub make_oid
-{
-  my $self = shift;
-  my $class_name = shift;
-  my @ids = @_;
-	
-  my $class_id = $self->class_id($class_name);
-  
-  my @oids = map {$self->combine_ids($_,$class_id)} @ids;
-  
-  if ( wantarray ) {
-	return @oids;
-  } else {
-	return $oids[0];
-  }
-}
-
-# Given a class' name and a row's ID (or more than one,)
-# loads the object(s) from the DB and returns them.
 sub import_object
 {
     my $self = shift;
@@ -915,7 +870,6 @@ sub import_object
     }
 }
 
-# XXX - not documented or tested by test suite
 sub dummy_object
 {
     my $self = shift;
@@ -980,7 +934,7 @@ sub welcome
     my ($self, $obj, $id) = @_;
     $self->{set_id}->($obj, $id);
 
-    weaken( $self->{objects}{$id} = $obj );
+    Tangram::weaken( $self->{objects}{$id} = $obj );
   }
 
 sub goodbye
@@ -991,7 +945,6 @@ sub goodbye
     delete $self->{PREFETCH}{$id};
   }
 
-# XXX - not documented or tested by test suite
 sub shrink
   {
     my ($self) = @_;
@@ -1013,51 +966,24 @@ sub read_object
 
     my $schema = $self->{schema};
 
-    my ($obj, $target, $is_dummy);
+    my $obj = $schema->{make_object}->($class);
 
-    if (exists $self->{objects}{$id} && defined $self->{objects}{$id}) {
-	# it's already in the cache, just return it.
-	$obj = $self->{objects}{$id};
-
-	# XXX - we are only doing this because we don't have an easy
-	# way of knowing how many columns each of the importers for
-	# this column type are returning.  It would be better to
-	# improve the importer protocol, and then just shift off the
-	# unneeded columns.
-
-	# the only reason we need to shift them off is for
-	# $cursor->residue(), which would otherwise return a variable
-	# number of items depending on whether we already had the row
-	# hot or not.
-
-	$target = bless {}, "dummy";
-	if ( $Tangram::TRACE ) {
-	    print $Tangram::TRACE __PACKAGE__.": made dummy object "
-		."$target\n";
-	}
-	$is_dummy = 1;
-    } else {
-	# do this only if object is not loaded yet
-	$obj = $schema->{make_object}->($class);
-	$self->welcome($obj, $id);
-	$target = $obj;
+    unless (exists $self->{objects}{$id} && defined $self->{objects}{$id}) {
+      # do this only if object is not loaded yet
+      # otherwise we're just skipping columns in $row
+      $self->welcome($obj, $id);
     }
 
-    _row_to_object($self, $target, $id, $class, $row, @parts);
-    CORE::bless $target, "dummy" if $is_dummy;
+    _row_to_object($self, $obj, $id, $class, $row, @parts);
+
     return $obj;
   }
-{
-package dummy;
-sub AUTOLOAD { }
-}
 
 sub _row_to_object
   {
     my ($self, $obj, $id, $class, $row) = @_;
 	my $context = { storage => $self, id => $id, layout1 => $self->{layout1} };
 	$self->{schema}->classdef($class)->get_importer($context)->($obj, $row, $context);
-    # XXX - not documented, probably badly named.
     if (my $x=$obj->can("T2_import")) {
 	$x->($obj);
     }
@@ -1113,7 +1039,6 @@ sub select {
 	return $cursor->select(@args);
   }
   
-  # XXX - not tested by test suite
   my ($first, @others) = @$target;
   
   my @cache = map { $self->select( $_, @args ) } @others;
@@ -1139,17 +1064,16 @@ sub select {
   return @results;
 }
 
-# XXX - not tested by test suite
 sub cursor_object
   {
     my ($self, $class) = @_;
-    $self->{IMPLICIT}{$class} ||= Tangram::Expr::RDBObject->new($self, $class)
+    $self->{IMPLICIT}{$class} ||= Tangram::RDBObject->new($self, $class)
 }
 
 sub query_objects
 {
     my ($self, @classes) = @_;
-    map { Tangram::Expr::QueryObject->new(Tangram::Expr::RDBObject->new($self, $_)) } @classes;
+    map { Tangram::QueryObject->new(Tangram::RDBObject->new($self, $_)) } @classes;
 }
 
 sub remote
@@ -1164,7 +1088,6 @@ sub expr
     return shift->expr( @_ );
   }
 
-# XXX - not tested by test suite
 sub object
 {
     carp "cannot be called in list context; use objects instead" if wantarray;
@@ -1173,39 +1096,81 @@ sub object
     $obj;
 }
 
-sub aggregate
-{
-    my $self = shift;
-    my $function = shift;
-    my $expr = shift;
-    my $filter = shift;
-    do {
-	$filter = $expr;
-	$expr = Tangram::Expr->new
-	    (Tangram::Type::Number->instance,
-	     '*', $filter->objects);
-    } if $expr->isa("Tangram::Expr::Filter");
-
-    my @data = $self->select(undef,
-			     ($filter ? (filter => $filter) : ()),
-			      retrieve => [ map { $_->$function() }
-						(ref ($expr) eq "ARRAY"
-						 ? @$expr : $expr) ],
-			    );
-
-    return $data[0]
-}
-
 sub count
 {
     my $self = shift;
-    $self->aggregate("count", @_);
+
+    my ($target, $filter);
+    my $objects = Set::Object->new;
+
+    if (@_ == 1)
+    {
+	$target = '*';
+	$filter = shift;
+    }
+    else
+    {
+	my $expr = shift;
+	$target = $expr->{expr};
+	$objects->insert($expr->objects);
+	$filter = shift;
+    }
+
+    my @filter_expr;
+
+    if ($filter)
+    {
+	$objects->insert($filter->objects);
+	@filter_expr = ( "($filter->{expr})" );
+    }
+
+    my $sql = "SELECT COUNT($target) FROM " . join(', ', map { $_->from } $objects->members);
+   
+    $sql .= "\nWHERE " . join(' AND ', @filter_expr, map { $_->where } $objects->members);
+
+    print $Tangram::TRACE ">-\n$sql\n...\n" if $Tangram::TRACE;
+
+    return ($self->{db}->selectrow_array($sql))[0];
 }
 
 sub sum
 {
-    my $self = shift;
-    $self->aggregate("sum", @_);
+    my ($self, $expr, $filter) = @_;
+
+    my $expr_is_array = ref($expr) eq 'ARRAY';
+
+    my $objects = Set::Object->new(
+				   $expr_is_array ? 
+				   map($_->objects, @$expr) :
+				   $expr->objects,
+				  );
+
+    my @filter_expr;
+
+    if ($filter)
+    {
+	$objects->insert($filter->objects);
+	@filter_expr = ( "($filter->{expr})" );
+    }
+
+    my $sql = "SELECT " .
+      join(', ',
+	   map("SUM($_->{expr})",
+	       $expr_is_array ? @$expr : ($expr),
+	      ),
+	  ) . " FROM " . join(', ', map { $_->from } $objects->members);
+
+    $sql .= "\nWHERE " . join(' AND ', @filter_expr, map { $_->where } $objects->members);
+
+    print $Tangram::TRACE ">-\n$sql\n...\n" if $Tangram::TRACE;
+
+    my @result = $self->{db}->selectrow_array($sql);
+
+    if ( $expr_is_array ) {
+      return wantarray ? @result : \@result;
+    } else {
+      return $result[0];
+    }
 }
 
 sub id
@@ -1277,7 +1242,6 @@ sub _kind_class_ids
     return @ids;
 }
 
-# XXX - not tested by test suite
 sub is_persistent
 {
     my ($self, $obj) = @_;
@@ -1328,7 +1292,7 @@ sub connect
 
     @$self{ -cs, -user, -pw } = ($cs, $user, $pw);
 
-    $self->{driver} = $opts->{driver} || Tangram::Relational->detect($cs);
+    $self->{driver} = $opts->{driver} || Tangram::Relational->new;
 
     my $db = $opts->{dbh};
     unless ( $db ) {
@@ -1387,7 +1351,6 @@ sub sql_do
 	  : croak $DBI::errstr;
 }
 
-# XXX - not tested by test suite
 sub sql_selectall_arrayref
 {
     my ($self, $sql, $dbh) = @_;
@@ -1440,7 +1403,6 @@ sub unload
     }
   }
 
-# XXX - not tested by test suite
 sub unload_all {
     my $self = shift;
     my $send_method = shift;
@@ -1480,10 +1442,8 @@ sub unload_all {
     #$self->SUPER::unload_all();
 }
 
-# XXX - not reached (?)
 sub from_dual { "" }
 
-# XXX - not tested by test suite
 sub ping {
     my $self = shift;
 
@@ -1504,7 +1464,6 @@ sub ping {
     #}
 }
 
-# XXX - not tested by test suite
 sub recycle {
     my $self = shift;
     my $send_method = shift;
@@ -1516,13 +1475,11 @@ sub recycle {
 	if $Tangram::TRACE;
 }
 
-# XXX - wtf?
 sub clear_stats {
     my $self = shift;
     $self->{stats} = undef;
 }
 
-# XXX - wtf?
 sub add_stat {
     my $self = shift;
     my $stat = shift;
@@ -1577,4 +1534,29 @@ sub DESTROY
     }
 }
 
+package Tangram::Storage::Statement;
+
+sub new
+{
+    my $class = shift;
+    bless { @_ }, $class;
+}
+
+sub fetchrow
+{
+    return shift->{statement}->fetchrow;
+}
+
+sub close
+{
+    my $self = shift;
+
+    if ($self->{storage})
+    {
+	$self->{statement}->finish;
+	$self->{storage}->close_connection($self->{connection});
+	%$self = ();
+    }
+}
+
 1;
