@@ -1,4 +1,4 @@
-
+# (c) Sound Object Logic 2000-2001
 
 use strict;
 use Tangram::Core;
@@ -55,11 +55,6 @@ sub drop_sequence_sql {
     return "DROP TABLE seq_$sequence_name";
 }
 
-sub limit_sql {
-    my $self = shift;
-    return (limit => shift);
-}
-
 package Tangram::mysql::Storage;
 
 use Tangram::Storage;
@@ -86,7 +81,7 @@ sub tx_start
     my $storage = shift;
     unless (@{ $storage->{tx} }) {
 	if ( $storage->{no_tx} ) {
-	    $storage->sql_do (q{SELECT GET_LOCK('tx', 10)} ); #})  #cperl-mode--
+	    $storage->sql_do (q{SELECT GET_LOCK("tx", 10)} ); #})  #cperl-mode--
 	}
     }
     $storage->SUPER::tx_start(@_);
@@ -98,7 +93,7 @@ sub tx_commit
     $storage->SUPER::tx_commit(@_);
     unless (@{ $storage->{tx} }) {
 	if ( $storage->{no_tx} ) {
-	    $storage->sql_do(q/SELECT RELEASE_LOCK('tx')/)
+	    $storage->sql_do(q/SELECT RELEASE_LOCK("tx")/)
 	}
     }
   }
@@ -107,7 +102,7 @@ sub tx_rollback
   {
     my $storage = shift;
     if ( $storage->{no_tx} ) {
-	$storage->sql_do(q/SELECT RELEASE_LOCK('tx')/);
+	$storage->sql_do(q/SELECT RELEASE_LOCK("tx")/);
     }
     $storage->SUPER::tx_rollback(@_);
   }
